#!/usr/bin/perl
# This program is free software, released under the Artistic license.
# Scott Cruzen <sic@lerp.com>

# Requires:
# Expect-1.10
# IO-Tty-0.04
# IO-Stty-.02
# TermReadKey-2.14

use Expect;
use Getopt::Long;
use Term::ReadKey;
use POSIX qw(strftime);

use Data::Dumper;

eval { require "names.pl"; };

warn $@ if $@;

#$Expect::Exp_Internal = 1;

my $usage = << "END";
mass.pl version 1.01
$0 --script script --machines machine1,...,machineN
--name namedlist --send file --retrieve file --ssh [1 or 2]
--noroot --sshpass --sshphrase --su --sudo --user username --bd --noping
--ssh [1 or 2]

general opts:
--script script\t\tspecify the name of the script to run on the targets
--machines machines\toperate on this comma separated list of machines
--name namedlist\tuse the named list of machines from names.pl
--send files\t\tsend this comma separated list of files
--retrieve files\tscp these files back after executing the script

auth opts:
--noroot\t\tdon't try to use su to become root
--sshpass\t\task for password to authenticate to ssh with
--sshphrase\t\task for passphrase to authenticate to ssh with
--su\t\t\tbecome root via su instead of sudo
--sudo\t\t\tbecome root via sudo instead of su
--user username\t\tspecify the user we should scp and ssh as
--bd\t\t\tbackdoor, equal to --noroot --user root --sshpass

misc opts:
--ssh option\t\tspecify ssh options (version etc)
--retonly\t\tonly retrieve files, no script
END

my $date = strftime "%Y%m%d", localtime;

my @send;
my @retrieve;
my @machines = ();
my $name;
my $user;
my $script;
my $retryct = 3;
my (@failed, @passed, @fatal);
my $ofh;
my $passwd;
my $noping;
my $retonly;
my $sshopts;
my $usesudo = 1;

#######################################################################
# process command line options

@args = @ARGV;

GetOptions( "script=s" => \$script, "machines=s" => \@machines,
"noroot" => \$noroot, "sudo" => \$usesudo, "su" => \$usesu,
"name=s" => \$name, "sshpass" => \$sshpass, "user=s" => \$user,
"send:s" => \@send, "retrieve:s" => \@retrieve, "bd" => \$bd,
"ssh:s" => \$sshopts, "noping" => \$noping, "sshphrase" => \$sshphrase,
"retonly" => \$retonly );

# get a list of machines from names.pl
@machines = eval { @$name } if ($name);

die $usage unless (($script || ($retonly && @retrieve)) && @machines);

unless ($user) { $user = $ENV{LOGNAME}; }

push @send, $script;

# check that $script starts with #!
open FILE, $script;
$shbang = <FILE>;
die "$script doesn't look like a shell script" if ($shbang !~ /^#!/);
$shbang =~ /^#!.*\/(.*)$/;
$prog = $1;

# scan the script for any '# send:' or '# retrieve:" lines
while (<FILE>) {
	/^\# send: (.*)/ && do { push @send, $1; };
	/^\# retrieve: (.*)/ && do { push @retrieve, $1; };
}
close FILE;

if ($bd) {
	$noroot = $sshpass = $bd;
	$user = "root";
}

if ($usesu) {
    $usesudo = 0;
}

@send = split(/,/,join(',',@send));
@retrieve = split(/,/,join(',',@retrieve));
@machines = split(/,/,join(',',@machines));

for (@send) {
	die "Couldn't find $_\n" unless -f;
}

#######################################################################
# get the root/sudo password
$passwd = getpass($usesudo ? "sudo password: " : "su password: ") unless $noroot;

$sshpasswd = getpass("ssh password: ") if $sshpass;
$sshpassphrase = getpass("ssh passphrase: ") if $sshphrase;

#######################################################################
# iterate over the specified machines

while (@machines && $retryct) {
	@failed = ();

	print "trying to run $script on ",scalar @machines, " machines\n";
	printnice(@machines);
	print "\n";

	for my $host (@machines) {
		print "$host\n\n";

		eval { dostuff($host, $script, $user, \@send, \@retrieve); };

		for ($@) {
			/^warning/i && do {
				print STDERR "\n\n$@\n";
				push @failed, $host;
				last;
			};

			/^fatal/i && do {
				print STDERR "\n\n$@\n";
				push @fatal, $host;
				last;
			};

			push @passed, $host;
		}
	}

	@machines = @failed;

	print scalar @passed, " passed: ";
	printnice(@passed);

	print scalar @failed, " failed: ";
	printnice(@failed);

	if (--$retryct && @failed) {
		print "sleeping for 5 seconds\n";
		sleep(5);
	}
}

# zero the root password in memory (pointless)
$passwd =~ s/./z/g;
$sshpasswd =~ s/./z/g;

print scalar @fatal, " fatal errors: ";
printnice(@fatal);

# Save the report to a file
open FILE, ">>report-$script";

$ofh = select(FILE);

print scalar localtime(),"\n";
print "args: ", join(' ', @args), "\n";

print scalar @passed, " passed: ";
printnice(@passed);

print scalar @failed, " failed: ";
printnice(@failed);

print scalar @fatal, " fatal errors: ";
printnice(@fatal);
print "\n";

select($ofh);

#######################################################################
# subroutines

# dostuff:
# ping a machine, if there's a response, scp the script specified on
# the command line to the machine (and any --send files), then ssh in,
# su to root, execute the script, close the ssh connection, then scp any
# files specified in --retrieve back and store them in retrieve/$name
#
# returns undef on success or an error string
sub dostuff {
	my ($name, $script, $user, $send, $retrieve) = @_;	
	my $file;

	# ping is suid (usually), so by using system we avoid having
	# to run mass.pl as root
	if (!$noping && system("ping -n -c 1 $name")) {
		die "Warning: $name doesn't respond to ping\n";
	}

    if ($retonly) {
        goto retrieveonly;
    }
#$Expect::Exp_Internal = 1;

	my @t;
	push @t, "scp";
	push @t, "-o $sshopts" if $sshopts;
	push @t, @$send, "$user\@$name:";

	my $scp = Expect->spawn(@t) ||
		die "Warning: scp failed on $name\n";

	$scp->expect(undef,
		[ 'Are you sure you want to continue ',
			sub {
				print $scp "yes\r"; exp_continue;
			},
		],
		[ 'Enter passphrase ',
			sub {
				$scp->log_group(undef);
				$scp->log_stdout(undef);
				print $scp "$sshpassphrase\r";
				$scp->expect(0, '');
				$scp->clear_accum();
				$scp->log_stdout(1);
				$scp->log_group(1);
				exp_continue;
			}
		],
		[ 'password',
			sub {
				$scp->log_group(undef);
				$scp->log_stdout(undef);
				print $scp "$sshpasswd\r";
				$scp->expect(0, '');
				$scp->clear_accum();
				$scp->log_stdout(1);
				$scp->log_group(1);

				exp_continue;
			}
		],
		[ 'Connection closed by remote host',
			sub {
				die "Warning: scp failed on $name\n";
			}
		],
		[ 'lost connection',
			sub {
				die "Warning: scp failed on $name\n";
			}
		]
	);

	$scp->hard_close();
# my $ssh = Expect->spawn("ssh", '-t', '-l', $user, $name, "sh") ||

	@t = qw(ssh -x);
	push @t, "-o $ssh" if $ssh;
	push @t, "-t", "-l", $user, $name, "sh";

#	print Dumper @t;

	my $ssh = Expect->spawn(@t) ||
		die "Warning: ssh failed on $name\n";

#$Expect::Exp_Internal = 1;

	$ssh->expect(120,
		[ 'Connection closed by remote host',
			sub {
				die "Warning: ssh failed on $name\n";
			}
		],
		[ 'lost connection',
			sub {
				die "Warning: ssh failed on $name\n";
			}
		],
		[ 'Error reading response length from authentication socket.',
			sub {
				die "Warning: ssh failed on $name\n";
			}
		],
		[ 'Enter passphrase ',
			sub {
				$ssh->log_group(undef);
				$ssh->log_stdout(undef);
				print $ssh "$sshpassphrase\r";
				$ssh->expect(0, '');
				$ssh->clear_accum();
				$ssh->log_stdout(1);
				$ssh->log_group(1);
				exp_continue;
			}
		],
		[ 'password:',
			sub {
				$ssh->log_group(undef);
				$ssh->log_stdout(undef);
				print $ssh "$sshpasswd\r";
				$ssh->expect(0, '');
				$ssh->clear_accum();
				$ssh->log_stdout(1);
				$ssh->log_group(1);
				exp_continue;
			}
		],
		[ '[#$] $',
# at this point we're logged in
			sub {
				print $ssh "PS1='\$ '; PATH=/usr/local/bin:".
					"/bin:/usr/bin:/usr/sbin:/usr/local/sbin:".
					"/sbin ;export PS1;export PATH\r";
				beroot($ssh, $name) unless $noroot;

#"chmod a+x $script ;".
				my $args = join ' ', @ARGV;
				print $ssh "$prog $script $args && ".
					"echo script\\ done || ".
					"echo script\\ failed\r";

				$ssh->clear_accum();
			}
		],
		[ 'timeout',
			sub {
				$ssh->hard_close();
				die "Warning: No shell prompt on $name\n";
			}
		]
	);

    my $err;

	$ssh->expect(undef,
		[ 'script done' ],
		[ 'script failed',
			sub {
                print "failed!\n";
                $err = "Fatal: Script $script failed on $name\n";
			}
		]
	);

# cleanup
	print $ssh "/bin/rm @send && echo removed\\ @send\r";

	$ssh->expect(undef,
		[ "removed @send" ]
	);

	$ssh->hard_close();

    die $err if $err;

	print "\n\n";

retrieveonly:	
	if (@$retrieve) {
		mkdir("retrieve", 0755) unless -d "retrieve";
		mkdir("retrieve/$name", 0755) unless -d "retrieve/$name";

        my @t;
        push @t, "scp";
#        push @t, "-v";
        push @t, "-o $sshopts" if $sshopts;
        push @t, "$user\@$name:@$retrieve";
        push @t, "retrieve/$name";

#print Dumper @t;

        my $scp = Expect->spawn(@t) ||
            die "Warning: scp failed on $name\n";

        $scp->expect(undef,
            [ 'Are you sure you want to continue ',
                sub {
                    print $scp "yes\r"; exp_continue;
                },
            ],
            [ 'Enter passphrase ',
                sub {
                    $scp->log_group(undef);
                    $scp->log_stdout(undef);
                    print $scp "$sshpassphrase\r";
                    $scp->expect(0, '');
                    $scp->clear_accum();
                    $scp->log_stdout(1);
                    $scp->log_group(1);
                    exp_continue;
                }
            ],
            [ 'password',
                sub {
                    $scp->log_group(undef);
                    $scp->log_stdout(undef);
                    print $scp "$sshpasswd\r";
                    $scp->expect(0, '');
                    $scp->clear_accum();
                    $scp->log_stdout(1);
                    $scp->log_group(1);

                    exp_continue;
                }
            ],
            [ 'Connection closed by remote host',
                sub {
                    die "Warning: scp failed on $name\n";
                }
            ],
            [ 'lost connection',
                sub {
                    die "Warning: scp failed on $name\n";
                }
            ]
        );

        $scp->hard_close();

#		if (system(qq(scp $name:"$file" retrieve/$name))) {
#			die "Fatal: scp of $file failed\n";
#		}
	}
}

sub beroot {
	my ($ssh, $name) = @_;

#$Expect::Exp_Internal = 1;

	$ssh->clear_accum();

	if ($usesudo) {
		print $ssh "sudo -K ; sudo sh\r";
	} else {
		print $ssh "su root -c sh\r";
	}

	$ssh->expect(undef,
		[ 'Password:',
			sub {
				print $ssh "$passwd\r";
			}
		]
	);

	$ssh->expect(undef,
		[ 'not found',
			sub {
				$ssh->hard_close();
				die("Fatal: ".($usesudo ? "sudo" : "su").
					" not found on $name\n");
			}
		],
		[ 'timeout',
			sub {
				$ssh->hard_close();
				die "Fatal: No root prompt on $name\n";
			}
		],
		[ 'Sorry|incorrect password|Password:',
			sub {
				$ssh->hard_close();
				die "Fatal: No root prompt on $name\n";
			}
		],
		[ '[$#] $',
			sub {
				print $ssh "PATH=/usr/local/bin:/bin".
				":/usr/bin:/usr/sbin:/bin/sbin:/sbin ; ".
				"export PATH\r".
				"if [ `id|cut -d ' ' -f 1` = 'uid=0(root)' ];".
				" then PS1='# '; fi\r";
			}
		]
	);
$Expect::Exp_Internal = 0;
}

# use format to wrap text
# takes an array as input and prints it comma separated
# for example: n123, n456, n789
sub printnice {
	my $list;

format Something =
^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ~~
$list
.

	# format line break characters
	$: = ',';
	# input record separator
	$/ = '';
	# format name
	$~ = 'Something';

	$list = join(',',sort(@_));
	if ($list) {
		write;
	} else {
		print "\n";
	}
}

sub getpass {
	my ($prompt) = @_;

	print STDERR $prompt;

	ReadMode 2; # Turn off echo

	my $pass = <STDIN>;
	chomp $pass;

	ReadMode 0; # Reset tty mode before exiting

	print STDERR "\n";

	return $pass;
}

# $Id: mass.pl,v 3.4 2004/02/06 00:07:03 sic Exp $
